//---------------------------------------------------------
//RG Basic Tools Collection
//Copyright © 2004 Ronald D. Grindle
//
//This file is part of the Rslt services.
//
//Rslt services is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.
//
//Rslt services is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with Rslt services.  If not, see <http://www.gnu.org/licenses/>.
//---------------------------------------------------------

package com.qanavigation.Basics;

import java.io.IOException;
import java.util.ArrayList;

/**
 * @author RDG
 * 
 * 'Rslt' is a base class for all results that an application 
 * might return.
 * Its purpose is to transport, attached to the original result, 
 * information about the state of the result.
 * Services can also be derived from result. The status of
 * the requested service is then inquirable  through the result
 * state of the object, rather then through return codes.
 * 
 * Due to the fact that the name 'Result' was unfortunatly arlready taken
 * by the Interface 'Result' for javax.xml.transform (very shortsighted decision)
 * and 'Outcome' was too clumsy for my taste I decided to abbreviate by 
 * dropping the vowels, thus resulting in 'Rslt'.
 */

public class Rslt implements IRslt {

	protected ArrayList<String> m_ResInfo = null;
	protected ArrayList<String> m_ResInitInfo = null;
	protected ArrayList<String> m_ResFinalInfo = null;	
	protected int m_ResStat;
	protected String m_Self;
	
	/**
	 * Default constructor. Status is "notInit",.
	 */
	public Rslt()	{
		
		setSelfReflect();
		m_ResStat = RsltStatVal.notInit;
	}
	
	/**
	 * Construct with default status 'OK' (true) or 'inError' (false).
	 * @param p_isOK
	 */
	public Rslt( boolean p_isOK ) {
		
		setSelfReflect();
		
		if( p_isOK ) {
			
			m_ResStat = RsltStatVal.isOK;
		}
		else {
			
		}
	}
	
	/**
	 * Constructor for a specific class
	 * @param p_c
	 */
	public Rslt( Class p_c ) {

		setSelfReflect( p_c ) ;
		m_ResStat = RsltStatVal.notInit;
	}
	
	/**
	 * Copy constructor for an Interface implementer.
	 * @param p_aRes
	 */
	public Rslt( IRslt p_aRes ) {

		setSelfReflect( p_aRes.getClass() ) ;
		setRsltStat( p_aRes );
	}

	/**
	 * Copy constructor for a Rslt object.
	 * @param p_aRes
	 */
	public Rslt( Rslt p_aRes ) {

		setSelfReflect( p_aRes.getClass() ) ;
		setRsltStat( p_aRes );
	}
	
	/**
	 * Is plugged together to 'p_Class + "." + p_Fnct + "()" + p_AddInf1 + " failed. " + p_AddInf2'.
	 * @param p_Class
	 * @param p_Fnct
	 * @param p_AddInf1
	 * @param p_AddInf2
	 */
	public void addRsltAddFailInfo( String p_Class, String p_Fnct, String p_AddInf1, String p_AddInf2 ) {
		
		//inf = p_Obj.getClass().getDeclaringClass().getName() + " "; Null Pointer
		//p_Obj.getClass().getName() get the top class
		//inf = p_Obj.getClass().getCanonicalName() + " "; not any better

		addRsltAddInfo( p_Class + "." + p_Fnct + "()" + p_AddInf1 + " failed. " + p_AddInf2 );
	}
	
	/**
	 * Is plugged together to 'p_Class + "." + p_Fnct + "()" + p_AddInf1 + " failed. " + p_AddInf2'.
	 * Appended only if notOK().
	 * @param p_Class
	 * @param p_Fnct
	 * @param p_AddInf1
	 * @param p_AddInf2
	 */
	public void addRsltAddFailInfoIf( String p_Class, String p_Fnct, String p_AddInf1, String p_AddInf2 ) {
		
		if( notOK() ) {
			
			addRsltAddFailInfo( p_Class, p_Fnct, p_AddInf1, p_AddInf2 );
		}
	}
	
	/* No proper class info
	 * Is plugged together to "p_Obj.getClass().getName() + "." + p_Fnct + "()" + p_AddInf + " failed.". 
	 * @param p_Obj
	 * @param p_Fnct
	 * @param p_AddInf
	 *
	public void addRsltAddFailInfo( Object p_Obj, String p_Fnct, String p_AddInf ) {
		
		addRsltAddFailInfo( p_Obj, p_Fnct, p_AddInf, "" );
	}	
	
	**
	 * Is plugged together to 'p_Obj.getClass().getName() + "." + p_Fnct + "()" + p_AddInf1 + " failed. " + p_AddInf2'.
	 * @param p_Obj
	 * @param p_Fnct
	 * @param p_AddInf1
	 * @param p_AddInf2
	 *
	public void addRsltAddFailInfo( Object p_Obj, String p_Fnct, String p_AddInf1, String p_AddInf2 ) {
		
		//addRsltAddInfo( p_Obj.getClass().getCanonicalName() );
		addRsltAddInfo( p_Obj.getClass().getName() + "." + p_Fnct + "()" + p_AddInf1 + " failed. " + p_AddInf2 );
	}
	*/
	
	// ------------------------------------------------------
	protected void addRsltAddInfo( ArrayList<String> p_Lst ) {
		
		if( p_Lst != null ) {
		
			if( p_Lst.size() > 0 ) {
				
				boolean isNew;
		
				isNew = createIfNewResInfo();	
				
				if( isNew ) {
					
					addRsltAddInfo( "Class: " + m_Self );
					addRsltAddInfo( "---" );
				}
				
				m_ResInfo.addAll( p_Lst );
			}
		}
	}
	
	// ------------------------------------------------------
	protected void addRsltFinalInfo( ArrayList<String> p_Lst ) {
		
		if( p_Lst != null ) {
		
			if( p_Lst.size() > 0 ) {
				
				if( m_ResFinalInfo == null ) {
					
					m_ResFinalInfo = new ArrayList<String>();
				}	
				
				m_ResFinalInfo.addAll( p_Lst );
			}
		}
	}

	// ------------------------------------------------------
	protected void addRsltInitialInfo( ArrayList<String> p_Lst ) {
		
		if( p_Lst != null ) {
		
			if( p_Lst.size() > 0 ) {
				
				if( m_ResInitInfo == null ) {
					
					m_ResInitInfo = new ArrayList<String>();
				}
				
				m_ResInitInfo.addAll( p_Lst );
			}
		}
	}
	
	/**
	 * Copies the result information from the other Rslt.
	 * @param  p_aRes
	 */
	public void addRsltAddInfo( IRslt p_aRes ) {
		
		ArrayList<String> lst;
		
		lst = p_aRes.getRsltInitialInfo();
		
		addRsltInitialInfo( lst );
		
		lst = p_aRes.getRsltAddInfo();
		
		addRsltAddInfo( lst );
		
		lst = p_aRes.getRsltFinalInfo();
		
		addRsltFinalInfo( lst );		
	}
		
	/**
	 * Add additional information to the Rslt. Makes it the initial info, if no initial info is available.
	 * @param p_AddInfo
	 */
	public void addRsltAddInfo( String p_AddInfo ) {
		
		if( p_AddInfo != null ) {

			if( p_AddInfo.length() > 0 ) {

				boolean isNew;

				isNew = createIfNewResInfo();

				if( isNew ) {

					m_ResInfo.add( "Class: " + m_Self );

					if( m_ResInitInfo == null ) {

						addRsltInitialInfo( p_AddInfo );
					}
				}
				
				m_ResInfo.add( p_AddInfo );
			}
		}
	}
	
	
	/**
	 * Add the final information to the Rslt.
	 * @param p_FinalInfo 
	 */
	public void addRsltFinalInfo( String p_FinalInfo ) {
		
		if( p_FinalInfo != null ) {
		
			if( p_FinalInfo.length() > 0 ) {
				
				if( m_ResFinalInfo == null ) {
					
					m_ResFinalInfo = new ArrayList<String>();
				}				
							
				m_ResFinalInfo.add( p_FinalInfo );
				
				addRsltAddInfo( p_FinalInfo );
			}
		}
	}	
	
	/**
	 * Add initial information to the Rslt.
	 * @param p_AddInfo 
	 */
	public void addRsltInitialInfo( String p_AddInfo ) {
		
		if( p_AddInfo != null ) {
		
			if( p_AddInfo.length() > 0 ) {
				
				if( m_ResInitInfo == null ) {
					
					m_ResInitInfo = new ArrayList<String>();
				}				
							
				m_ResInitInfo.add( p_AddInfo );
				
				addRsltAddInfo( p_AddInfo );
			}
		}
	}		
	
	/**
	 * Checks the status of the given Rslt and clones the information if its not OK.
	 * <pre>Simplifies code sequences like
	 *   "if( m_DB.notOK() ) {
	 *	    res.setResStat( m_DB );
	 *	  }
	 *  
	 *	  if( res.isOK() ) { ..." 
	 * to
	 *   "if( res.checkRes( m_DB ) ) {... "</pre> 
	 * @param p_Res
	 * @return isOK(), which also may be the state before the call! 
	 */
	public boolean checkRslt( IRslt p_Res ) {

		if( p_Res.notOK() ) {

			setRsltStat( p_Res );
		}

		return isOK();
	}
	
	/**
	 * Same as checkRslt( IRslt p_Res ) but tolerates if the result or itself isEmpty.
	 * @param p_Res
	 * @param p_EmptyIsOK 
	 * @return isOK(), which also may be the state before the call!  
	 */
	public boolean checkRslt( IRslt p_Res, boolean p_EmptyIsOK ) {

		if( p_Res.notOK( p_EmptyIsOK ) ) {

			setRsltStat( p_Res );
		}

		return isOK( p_EmptyIsOK );
	}
	
	/**
	 * <pre>Simplify 
	 *	"if( m_DB.notOK() ) {
	 *	   setResStat( m_DB );
	 *	 }
	 *	 if( isOK() ) { ..." 
	 * to
	 *	"if( checkRes( m_DB ) ) {... "</pre> 
	 * @param p_Res
	 * @return isOK(), which also may be the state before the call! 
	 */
	public boolean checkRslt( Rslt p_Res ) {

		if( p_Res.notOK() ) {

			setRsltStat( p_Res );
		}

		return isOK();
	}
	
	/**
	 * @param p_Res
	 * @param p_EmptyIsOK
	 */
	public boolean checkRslt( Rslt p_Res, boolean p_EmptyIsOK ) {

		if( p_Res.notOK( p_EmptyIsOK ) ) {

			setRsltStat( p_Res );
		}

		return isOK();
	}

	// ------------------------------------------------------
	protected boolean createIfNewResInfo() {
		
		boolean isNew = false;
		
		if( m_ResInfo == null ) {
			
			m_ResInfo = new ArrayList<String>();
			
			isNew = true;
		}
		
		return isNew;
	}
	
	/**
	 * @return
	 */
	public IRslt getRslt() {

		Rslt aRes = new Rslt();

		aRes.setRsltStat( this );

		return aRes;
	}

	/**
	 * @return
	 */
	public ArrayList<String> getRsltAddInfo() {
		
		boolean isNew;
		
		isNew = createIfNewResInfo();
		
		if( isNew ) {

			addRsltAddInfo( "Class: " + m_Self );
			addRsltAddInfo( "---" );
		}

		return m_ResInfo;
	}
	
	/**
	 * @return
	 */
	public ArrayList<String> getRsltFinalInfo() {
		
		if( m_ResFinalInfo == null ) {
			
			m_ResFinalInfo = new ArrayList<String>();
		}

		return m_ResFinalInfo;
	}
	
	/**
	 * @return
	 */
	public ArrayList<String> getRsltInitialInfo() {
		
		if( m_ResInitInfo == null ) {
			
			m_ResInitInfo = new ArrayList<String>();
		}

		return m_ResInitInfo;
	}	
	
	/**
	 * Returns the first of the messages on the message stack.
	 * @return
	 */
	public String getRsltInitialMsg( ) {

		String msg;

		msg = "";

		if( m_ResInitInfo != null ) {

			int x = m_ResInitInfo.size();

			if( x > 0 ) {

				msg = (String)m_ResInitInfo.get( 0 );
			}
		}

		return msg;
	}

	/**
	 * Returns the last final message. If none is found it returns the last message on the message stack.
	 * @return
	 */
	public String getRsltFinalMsg( ) {
		
		boolean notFound;

		String lastMsg;

		lastMsg = "";
		
		notFound = true;

		if( m_ResFinalInfo != null ) {

			int x = m_ResFinalInfo.size() -1;

			if( x > -1 ) {

				lastMsg = m_ResFinalInfo.get( x );
				
				notFound = false;
			}
		}
		
		if( notFound ) {
			
			if( m_ResInfo != null ) {

				int x = m_ResInfo.size() -1;

				if( x > -1 ) {

					lastMsg = m_ResInfo.get( x );
				}
			}
		}

		return lastMsg;
	}
	
	/**
	 * @return
	 */
	public int getRsltStatVal() {
		
		return m_ResStat;
	}
	
	/**
	 * @return
	 */
	public String getRsltStatValLbl() {
		
		return RsltStatVal.label[ m_ResStat ];
	}
	
	// ------------------------------------------------------
	protected void init() {

	}
	
	/**
	 * Check for the requested status.
	 * @param p_Stat
	 * @return 
	 */
	public boolean isInStat( int p_Stat ) {

		boolean isStat = false;

		if( m_ResStat == p_Stat ) {

			isStat = true;
		}

		return isStat;
	}

	/**
	 * If it is RsltStatVal.isOK it's OK;
	 * @return
	 */
	public boolean isOK() {

		boolean isOK = true;

		if( m_ResStat != RsltStatVal.isOK ) {

			isOK = false;
		}

		return isOK;
	}
	
	/**
	 * 
	 * @param p_EmptyIsOK
	 * @return
	 */
	public boolean isOK( boolean p_EmptyIsOK ) {

		boolean isOK;
	
		if( (m_ResStat == RsltStatVal.isEmpty) && p_EmptyIsOK ) {
				
				isOK = true;
		}
		else {
			
			isOK = isOK();
		}

		return isOK;
	}

	/**
	 * Anything that is not isOK() is not OK;
	 * @return
	 */
	public boolean notOK() {
		
		boolean notOK = !(isOK());

		return notOK;
	}
	
	/**
	 * 
	 * @param p_EmptyIsOK
	 * @return
	 */
	public boolean notOK( boolean p_EmptyIsOK ) {

		boolean notOK = !(isOK( p_EmptyIsOK ));

		return notOK;
	}
	
	/**
	 * Set back to OK, dumps all messages.
	 */
	public void resetRslt( ) {

		if( null != m_ResInfo ) {
		
			m_ResInfo.clear();	
		}
			
		setOK();
	}
	
	/**
	 * Sets the status to "In Error" and adds the messages to the list of messages.
	 * @param p_ErrMsg
	 */
	public void setInError(String p_ErrMsg) {

		setRsltStat( RsltStatVal.inError, p_ErrMsg );
	}

	/**
	 *
	 */
	public void setOK() {

		m_ResStat = RsltStatVal.isOK;
	}
	
	/**
	 * Sets the Rslt in Error and collects the information from the exception.
	 * @param p_Excp
	 */
	public void setRsltByExcp(Exception p_Excp) {
	
		setRsltByExcp( (Throwable)p_Excp );
	}
	
	/**
	 * Sets the Rslt in Error and collects the information from the exception.
	 * @param p_Excp
	 */
	public void setRsltByExcp(Throwable p_Excp) {
		
		int i, x;
		Throwable cause;
		StackTraceElement[] ste;
		String msg;
		
		m_ResStat = RsltStatVal.inError;
		
		msg = p_Excp.getMessage();
		
		if( null != msg ) {
			
			addRsltAddInfo( "Msg: '" +  msg + "'" );
		}
		
		addRsltAddInfo( "Excp: '" + p_Excp.toString() + "'" );
	
		cause = p_Excp.getCause();
		
		if( null != cause ) {
			
			addRsltAddInfo( "Cause: '" + cause.toString() + "'" );
		}
		
		ste = p_Excp.getStackTrace();
		
		x = ste.length;
		
		if( x > 0 )  {
			
			addRsltAddInfo( "Stack Trace:" );
		}
		
		for( i = 0; i < x; ++i ) {
			
			addRsltAddInfo( "  " + ste[i].toString() );
		}		
	}
	
	/**
	 * Sets the Rslt in Error and collects the information from the exception.
	 * @param p_Excp
	 */
	public void setRsltByExcp( IOException p_Excp ) {
		
		setRsltByExcp( (Throwable)p_Excp );
	}
	
	/**
	 * Sets the status to the given state.
	 * @param p_aResStat
	 */
	public void setRsltStat(int p_aResStat) {
			
		m_ResStat = p_aResStat;
	}

	/**
	 * Set Rslt status by status value, with message.
	 * @param p_aResStat
	 * @param p_ErrMsg
	 */
	public void setRsltStat( int p_aResStat, String p_ErrMsg ) {
			
		m_ResStat = p_aResStat;
		addRsltAddInfo( p_ErrMsg );
	}
	
	/**
	 * Set Rslt status from another Rslt.
	 * @param p_aRes
	 */
	public void setRsltStat(IRslt p_aRes) {

		m_ResStat = p_aRes.getRsltStatVal();
		
		addRsltAddInfo( p_aRes );
	}

	/**
	 * Set Rslt status from another Rslt with message.
	 * @param p_aRes
	 * @param p_ErrMsg
	 */
	public void setRsltStat(IRslt p_aRes, String p_ErrMsg ) {

		setRsltStat( p_aRes );
		addRsltAddInfo( p_ErrMsg  );
	}
	
	/**
	 * Gets information about the class.
	 * Interface implementation: make a copy of this method!
	 */
	public void setSelfReflect() {

		setSelfReflect( this.getClass() );	
	}
	
	// The actual worker method
	protected  void setSelfReflect( Class  p_c ) {

		//sRefl = c.getPackage().getName();
		m_Self = p_c.getName();
	}

}
